import dataclasses as dc
import typing as ta

from omlish.lite.check import check


DeployPathKind = ta.Literal['dir', 'file']  # ta.TypeAlias
DeployPathPlaceholder = ta.Literal['app', 'tag', 'conf']  # ta.TypeAlias


##


DeployHome = ta.NewType('DeployHome', str)

DeployApp = ta.NewType('DeployApp', str)
DeployTag = ta.NewType('DeployTag', str)
DeployRev = ta.NewType('DeployRev', str)
DeployKey = ta.NewType('DeployKey', str)


##


@dc.dataclass(frozen=True)
class DeployAppTag:
    app: DeployApp
    tag: DeployTag

    def __post_init__(self) -> None:
        for s in [self.app, self.tag]:
            check.non_empty_str(s)
            check.equal(s, s.strip())

    def placeholders(self) -> ta.Mapping[DeployPathPlaceholder, str]:
        return {
            'app': self.app,
            'tag': self.tag,
        }



# ruff: noqa: UP006 UP007
import typing as ta

from omlish.lite.cached import cached_nullary
from omlish.lite.check import check
from omlish.os.atomics import AtomicPathSwap
from omlish.os.atomics import AtomicPathSwapKind
from omlish.os.atomics import AtomicPathSwapping
from omlish.os.atomics import TempDirAtomicPathSwapping

from .paths.owners import SingleDirDeployPathOwner
from .types import DeployHome


class DeployTmpManager(
    SingleDirDeployPathOwner,
    AtomicPathSwapping,
):
    def __init__(
            self,
            *,
            deploy_home: ta.Optional[DeployHome] = None,
    ) -> None:
        super().__init__(
            owned_dir='tmp',
            deploy_home=deploy_home,
        )

    @cached_nullary
    def _swapping(self) -> AtomicPathSwapping:
        return TempDirAtomicPathSwapping(
            temp_dir=self._make_dir(),
            root_dir=check.non_empty_str(self._deploy_home),
        )

    def begin_atomic_path_swap(
            self,
            kind: AtomicPathSwapKind,
            dst_path: str,
            **kwargs: ta.Any,
    ) -> AtomicPathSwap:
        return self._swapping().begin_atomic_path_swap(
            kind,
            dst_path,
            **kwargs,
        )



# ruff: noqa: UP006 UP007
import datetime
import os.path
import shutil
import typing as ta

from omlish.lite.cached import cached_nullary
from omlish.lite.check import check
from omlish.os.paths import relative_symlink

from .conf import DeployConfManager
from .git import DeployGitManager
from .paths.owners import DeployPathOwner
from .paths.paths import DeployPath
from .specs import DeploySpec
from .types import DeployAppTag
from .types import DeployHome
from .types import DeployKey
from .types import DeployRev
from .types import DeployTag
from .venvs import DeployVenvManager


def make_deploy_tag(
        rev: DeployRev,
        key: DeployKey,
        *,
        utcnow: ta.Optional[datetime.datetime] = None,
) -> DeployTag:
    if utcnow is None:
        utcnow = datetime.datetime.now(tz=datetime.timezone.utc)  # noqa
    now_fmt = '%Y%m%dT%H%M%SZ'
    now_str = utcnow.strftime(now_fmt)
    return DeployTag('-'.join([now_str, rev, key]))


class DeployAppManager(DeployPathOwner):
    def __init__(
            self,
            *,
            deploy_home: ta.Optional[DeployHome] = None,

            conf: DeployConfManager,
            git: DeployGitManager,
            venvs: DeployVenvManager,
    ) -> None:
        super().__init__()

        self._deploy_home = deploy_home

        self._conf = conf
        self._git = git
        self._venvs = venvs

    #

    _APP_TAG_DIR_STR = 'tags/apps/@app/@tag/'
    _APP_TAG_DIR = DeployPath.parse(_APP_TAG_DIR_STR)

    _CONF_TAG_DIR_STR = 'tags/conf/@tag--@app/'
    _CONF_TAG_DIR = DeployPath.parse(_CONF_TAG_DIR_STR)

    _DEPLOY_DIR_STR = 'deploys/@tag--@app/'
    _DEPLOY_DIR = DeployPath.parse(_DEPLOY_DIR_STR)

    _APP_DEPLOY_LINK = DeployPath.parse(f'{_DEPLOY_DIR_STR}apps/@app')
    _CONF_DEPLOY_LINK = DeployPath.parse(f'{_DEPLOY_DIR_STR}conf')

    @cached_nullary
    def get_owned_deploy_paths(self) -> ta.AbstractSet[DeployPath]:
        return {
            self._APP_TAG_DIR,

            self._CONF_TAG_DIR,

            self._DEPLOY_DIR,

            self._APP_DEPLOY_LINK,
            self._CONF_DEPLOY_LINK,

            *[
                DeployPath.parse(f'{self._APP_TAG_DIR_STR}{sfx}/')
                for sfx in [
                    'conf',
                    'git',
                    'venv',
                ]
            ],
        }

    #

    async def prepare_app(
            self,
            spec: DeploySpec,
    ) -> None:
        app_tag = DeployAppTag(spec.app, make_deploy_tag(spec.git.rev, spec.key()))

        #

        deploy_home = check.non_empty_str(self._deploy_home)

        def build_path(pth: DeployPath) -> str:
            return os.path.join(deploy_home, pth.render(app_tag.placeholders()))

        app_tag_dir = build_path(self._APP_TAG_DIR)
        conf_tag_dir = build_path(self._CONF_TAG_DIR)
        deploy_dir = build_path(self._DEPLOY_DIR)
        app_deploy_link = build_path(self._APP_DEPLOY_LINK)
        conf_deploy_link_file = build_path(self._CONF_DEPLOY_LINK)

        #

        os.makedirs(deploy_dir)

        deploying_link = os.path.join(deploy_home, 'deploys/deploying')
        relative_symlink(
            deploy_dir,
            deploying_link,
            target_is_directory=True,
            make_dirs=True,
        )

        #

        os.makedirs(app_tag_dir)
        relative_symlink(
            app_tag_dir,
            app_deploy_link,
            target_is_directory=True,
            make_dirs=True,
        )

        #

        os.makedirs(conf_tag_dir)
        relative_symlink(
            conf_tag_dir,
            conf_deploy_link_file,
            target_is_directory=True,
            make_dirs=True,
        )

        #

        def mirror_symlinks(src: str, dst: str) -> None:
            def mirror_link(lp: str) -> None:
                check.state(os.path.islink(lp))
                shutil.copy2(
                    lp,
                    os.path.join(dst, os.path.relpath(lp, src)),
                    follow_symlinks=False,
                )

            for dp, dns, fns in os.walk(src, followlinks=False):
                for fn in fns:
                    mirror_link(os.path.join(dp, fn))

                for dn in dns:
                    dp2 = os.path.join(dp, dn)
                    if os.path.islink(dp2):
                        mirror_link(dp2)
                    else:
                        os.makedirs(os.path.join(dst, os.path.relpath(dp2, src)))

        current_link = os.path.join(deploy_home, 'deploys/current')
        # if os.path.exists(current_link):
        #     mirror_symlinks(
        #         os.path.join(current_link, 'conf'),
        #         conf_tag_dir,
        #     )
        #     mirror_symlinks(
        #         os.path.join(current_link, 'apps'),
        #         os.path.join(deploy_dir, 'apps'),
        #     )

        #

        git_dir = os.path.join(app_tag_dir, 'git')
        await self._git.checkout(
            spec.git,
            git_dir,
        )

        #

        if spec.venv is not None:
            venv_dir = os.path.join(app_tag_dir, 'venv')
            await self._venvs.setup_venv(
                spec.venv,
                git_dir,
                venv_dir,
            )

        #

        if spec.conf is not None:
            conf_dir = os.path.join(app_tag_dir, 'conf')
            await self._conf.write_conf(
                spec.conf,
                app_tag,
                conf_dir,
                conf_tag_dir,
            )

        #

        os.replace(deploying_link, current_link)



# ruff: noqa: UP006 UP007
"""
TODO:
 - 'repos'?

git/github.com/wrmsr/omlish <- bootstrap repo
 - shallow clone off bootstrap into /apps

github.com/wrmsr/omlish@rev
"""
import functools
import os.path
import typing as ta

from omlish.asyncs.asyncio.subprocesses import asyncio_subprocesses
from omlish.lite.cached import async_cached_nullary
from omlish.lite.check import check
from omlish.os.atomics import AtomicPathSwapping

from .paths.owners import SingleDirDeployPathOwner
from .specs import DeployGitRepo
from .specs import DeployGitSpec
from .types import DeployHome
from .types import DeployRev


##


class DeployGitManager(SingleDirDeployPathOwner):
    def __init__(
            self,
            *,
            deploy_home: ta.Optional[DeployHome] = None,
            atomics: AtomicPathSwapping,
    ) -> None:
        super().__init__(
            owned_dir='git',
            deploy_home=deploy_home,
        )

        self._atomics = atomics

        self._repo_dirs: ta.Dict[DeployGitRepo, DeployGitManager.RepoDir] = {}

    class RepoDir:
        def __init__(
                self,
                git: 'DeployGitManager',
                repo: DeployGitRepo,
        ) -> None:
            super().__init__()

            self._git = git
            self._repo = repo
            self._dir = os.path.join(
                self._git._make_dir(),  # noqa
                check.non_empty_str(repo.host),
                check.non_empty_str(repo.path),
            )

        @property
        def repo(self) -> DeployGitRepo:
            return self._repo

        @property
        def url(self) -> str:
            if self._repo.username is not None:
                return f'{self._repo.username}@{self._repo.host}:{self._repo.path}'
            else:
                return f'https://{self._repo.host}/{self._repo.path}'

        #

        async def _call(self, *cmd: str) -> None:
            await asyncio_subprocesses.check_call(
                *cmd,
                cwd=self._dir,
            )

        #

        @async_cached_nullary
        async def init(self) -> None:
            os.makedirs(self._dir, exist_ok=True)
            if os.path.exists(os.path.join(self._dir, '.git')):
                return

            await self._call('git', 'init')
            await self._call('git', 'remote', 'add', 'origin', self.url)

        async def fetch(self, rev: DeployRev) -> None:
            await self.init()
            await self._call('git', 'fetch', '--depth=1', 'origin', rev)

        #

        async def checkout(self, spec: DeployGitSpec, dst_dir: str) -> None:
            check.state(not os.path.exists(dst_dir))
            with self._git._atomics.begin_atomic_path_swap(  # noqa
                    'dir',
                    dst_dir,
                    auto_commit=True,
                    make_dirs=True,
            ) as dst_swap:
                await self.fetch(spec.rev)

                dst_call = functools.partial(asyncio_subprocesses.check_call, cwd=dst_swap.tmp_path)
                await dst_call('git', 'init')

                await dst_call('git', 'remote', 'add', 'local', self._dir)
                await dst_call('git', 'fetch', '--depth=1', 'local', spec.rev)
                await dst_call('git', 'checkout', spec.rev, *(spec.subtrees or []))

    def get_repo_dir(self, repo: DeployGitRepo) -> RepoDir:
        try:
            return self._repo_dirs[repo]
        except KeyError:
            repo_dir = self._repo_dirs[repo] = DeployGitManager.RepoDir(self, repo)
            return repo_dir

    async def checkout(
            self,
            spec: DeployGitSpec,
            dst_dir: str,
    ) -> None:
        await self.get_repo_dir(spec.repo).checkout(spec, dst_dir)



# ruff: noqa: UP006 UP007
"""
TODO:
 - @conf DeployPathPlaceholder? :|
 - post-deploy: remove any dir_links not present in new spec
  - * only if succeeded * - otherwise, remove any dir_links present in new spec but not previously present?
   - no such thing as 'previously present'.. build a 'deploy state' and pass it back?
 - ** whole thing can be atomic **
  - 1) new atomic temp dir
  - 2) for each subdir not needing modification, hardlink into temp dir
  - 3) for each subdir needing modification, new subdir, hardlink all files not needing modification
  - 4) write (or if deleting, omit) new files
  - 5) swap top level
 - ** whole deploy can be atomic(-ish) - do this for everything **
  - just a '/deploy/current' dir
  - some things (venvs) cannot be moved, thus the /deploy/venvs dir
  - ** ensure (enforce) equivalent relpath nesting
"""
import os.path
import typing as ta

from omlish.lite.check import check
from omlish.os.paths import is_path_in_dir
from omlish.os.paths import relative_symlink

from .paths.paths import DEPLOY_PATH_PLACEHOLDER_SEPARATOR
from .specs import AppDeployConfLink
from .specs import DeployConfFile
from .specs import DeployConfLink
from .specs import DeployConfSpec
from .specs import TagDeployConfLink
from .types import DeployAppTag
from .types import DeployHome


class DeployConfManager:
    def __init__(
            self,
            *,
            deploy_home: ta.Optional[DeployHome] = None,
    ) -> None:
        super().__init__()

        self._deploy_home = deploy_home

    #

    async def _write_conf_file(
            self,
            cf: DeployConfFile,
            conf_dir: str,
    ) -> None:
        conf_file = os.path.join(conf_dir, cf.path)
        check.arg(is_path_in_dir(conf_dir, conf_file))

        os.makedirs(os.path.dirname(conf_file), exist_ok=True)

        with open(conf_file, 'w') as f:  # noqa
            f.write(cf.body)

    #

    class _ComputedConfLink(ta.NamedTuple):
        is_dir: bool
        link_src: str
        link_dst: str

    def _compute_conf_link_dst(
            self,
            link: DeployConfLink,
            app_tag: DeployAppTag,
            conf_dir: str,
            link_dir: str,
    ) -> _ComputedConfLink:
        link_src = os.path.join(conf_dir, link.src)
        check.arg(is_path_in_dir(conf_dir, link_src))

        #

        if (is_dir := link.src.endswith('/')):
            # @conf/ - links a directory in root of app conf dir to conf/@conf/@dst/
            check.arg(link.src.count('/') == 1)
            link_dst_pfx = link.src
            link_dst_sfx = ''

        elif '/' in link.src:
            # @conf/file - links a single file in a single subdir to conf/@conf/@dst--file
            d, f = os.path.split(link.src)
            # TODO: check filename :|
            link_dst_pfx = d + '/'
            link_dst_sfx = DEPLOY_PATH_PLACEHOLDER_SEPARATOR + f

        else:  # noqa
            # @conf(.ext)* - links a single file in root of app conf dir to conf/@conf/@dst(.ext)*
            if '.' in link.src:
                l, _, r = link.src.partition('.')
                link_dst_pfx = l + '/'
                link_dst_sfx = '.' + r
            else:
                link_dst_pfx = link.src + '/'
                link_dst_sfx = ''

        #

        if isinstance(link, AppDeployConfLink):
            link_dst_mid = str(app_tag.app)
        elif isinstance(link, TagDeployConfLink):
            link_dst_mid = DEPLOY_PATH_PLACEHOLDER_SEPARATOR.join([app_tag.app, app_tag.tag])
        else:
            raise TypeError(link)

        #

        link_dst_name = ''.join([
            link_dst_pfx,
            link_dst_mid,
            link_dst_sfx,
        ])
        link_dst = os.path.join(link_dir, link_dst_name)

        return DeployConfManager._ComputedConfLink(
            is_dir=is_dir,
            link_src=link_src,
            link_dst=link_dst,
        )

    async def _make_conf_link(
            self,
            link: DeployConfLink,
            app_tag: DeployAppTag,
            conf_dir: str,
            link_dir: str,
    ) -> None:
        comp = self._compute_conf_link_dst(
            link,
            app_tag,
            conf_dir,
            link_dir,
        )

        #

        check.arg(is_path_in_dir(conf_dir, comp.link_src))
        check.arg(is_path_in_dir(link_dir, comp.link_dst))

        if comp.is_dir:
            check.arg(os.path.isdir(comp.link_src))
        else:
            check.arg(os.path.isfile(comp.link_src))

        #

        relative_symlink(  # noqa
            comp.link_src,
            comp.link_dst,
            target_is_directory=comp.is_dir,
            make_dirs=True,
        )

    #

    async def write_conf(
            self,
            spec: DeployConfSpec,
            app_tag: DeployAppTag,
            conf_dir: str,
            link_dir: str,
    ) -> None:
        for cf in spec.files or []:
            await self._write_conf_file(
                cf,
                conf_dir,
            )

        #

        for link in spec.links or []:
            await self._make_conf_link(
                link,
                app_tag,
                conf_dir,
                link_dir,
            )



# ruff: noqa: UP006 UP007
"""
TODO:
 - run/{.pid,.sock}
 - logs/...
 - current symlink
 - conf/{nginx,supervisor}
 - env/?
 - apps/<app>/shared
"""
import abc
import dataclasses as dc
import itertools
import typing as ta

from omlish.lite.check import check
from omlish.lite.strings import split_keep_delimiter

from ..types import DeployPathKind
from ..types import DeployPathPlaceholder


##


DEPLOY_PATH_PLACEHOLDER_SIGIL = '@'
DEPLOY_PATH_PLACEHOLDER_SEPARATOR = '--'

DEPLOY_PATH_PLACEHOLDER_DELIMITERS: ta.AbstractSet[str] = frozenset([
    DEPLOY_PATH_PLACEHOLDER_SEPARATOR,
    '.',
])

DEPLOY_PATH_PLACEHOLDERS: ta.FrozenSet[str] = frozenset([
    'app',
    'tag',
    'conf',
])


class DeployPathError(Exception):
    pass


class DeployPathRenderable(abc.ABC):
    @abc.abstractmethod
    def render(self, placeholders: ta.Optional[ta.Mapping[DeployPathPlaceholder, str]] = None) -> str:
        raise NotImplementedError


##


class DeployPathNamePart(DeployPathRenderable, abc.ABC):
    @classmethod
    def parse(cls, s: str) -> 'DeployPathNamePart':
        check.non_empty_str(s)
        if s.startswith(DEPLOY_PATH_PLACEHOLDER_SIGIL):
            return PlaceholderDeployPathNamePart(s[1:])
        elif s in DEPLOY_PATH_PLACEHOLDER_DELIMITERS:
            return DelimiterDeployPathNamePart(s)
        else:
            return ConstDeployPathNamePart(s)


@dc.dataclass(frozen=True)
class PlaceholderDeployPathNamePart(DeployPathNamePart):
    placeholder: str  # DeployPathPlaceholder

    def __post_init__(self) -> None:
        check.in_(self.placeholder, DEPLOY_PATH_PLACEHOLDERS)

    def render(self, placeholders: ta.Optional[ta.Mapping[DeployPathPlaceholder, str]] = None) -> str:
        if placeholders is not None:
            return placeholders[self.placeholder]  # type: ignore
        else:
            return DEPLOY_PATH_PLACEHOLDER_SIGIL + self.placeholder


@dc.dataclass(frozen=True)
class DelimiterDeployPathNamePart(DeployPathNamePart):
    delimiter: str

    def __post_init__(self) -> None:
        check.in_(self.delimiter, DEPLOY_PATH_PLACEHOLDER_DELIMITERS)

    def render(self, placeholders: ta.Optional[ta.Mapping[DeployPathPlaceholder, str]] = None) -> str:
        return self.delimiter


@dc.dataclass(frozen=True)
class ConstDeployPathNamePart(DeployPathNamePart):
    const: str

    def __post_init__(self) -> None:
        check.non_empty_str(self.const)
        for c in [*DEPLOY_PATH_PLACEHOLDER_DELIMITERS, DEPLOY_PATH_PLACEHOLDER_SIGIL, '/']:
            check.not_in(c, self.const)

    def render(self, placeholders: ta.Optional[ta.Mapping[DeployPathPlaceholder, str]] = None) -> str:
        return self.const


@dc.dataclass(frozen=True)
class DeployPathName(DeployPathRenderable):
    parts: ta.Sequence[DeployPathNamePart]

    def __post_init__(self) -> None:
        hash(self)
        check.not_empty(self.parts)
        for k, g in itertools.groupby(self.parts, type):
            if len(gl := list(g)) > 1:
                raise DeployPathError(f'May not have consecutive path name part types: {k} {gl}')

    def render(self, placeholders: ta.Optional[ta.Mapping[DeployPathPlaceholder, str]] = None) -> str:
        return ''.join(p.render(placeholders) for p in self.parts)

    @classmethod
    def parse(cls, s: str) -> 'DeployPathName':
        check.non_empty_str(s)
        check.not_in('/', s)

        i = 0
        ps = []
        while i < len(s):
            ns = [(n, d) for d in DEPLOY_PATH_PLACEHOLDER_DELIMITERS if (n := s.find(d, i)) >= 0]
            if not ns:
                ps.append(s[i:])
                break
            n, d = min(ns)
            ps.append(check.non_empty_str(s[i:n]))
            ps.append(s[n:n + len(d)])
            i = n + len(d)

        return cls(tuple(DeployPathNamePart.parse(p) for p in ps))


##


@dc.dataclass(frozen=True)
class DeployPathPart(DeployPathRenderable, abc.ABC):  # noqa
    name: DeployPathName

    @property
    @abc.abstractmethod
    def kind(self) -> DeployPathKind:
        raise NotImplementedError

    def render(self, placeholders: ta.Optional[ta.Mapping[DeployPathPlaceholder, str]] = None) -> str:
        return self.name.render(placeholders) + ('/' if self.kind == 'dir' else '')

    @classmethod
    def parse(cls, s: str) -> 'DeployPathPart':
        if (is_dir := s.endswith('/')):
            s = s[:-1]
        check.non_empty_str(s)
        check.not_in('/', s)

        n = DeployPathName.parse(s)
        if is_dir:
            return DirDeployPathPart(n)
        else:
            return FileDeployPathPart(n)


class DirDeployPathPart(DeployPathPart):
    @property
    def kind(self) -> DeployPathKind:
        return 'dir'


class FileDeployPathPart(DeployPathPart):
    @property
    def kind(self) -> DeployPathKind:
        return 'file'


#


@dc.dataclass(frozen=True)
class DeployPath:
    parts: ta.Sequence[DeployPathPart]

    @property
    def name_parts(self) -> ta.Iterator[DeployPathNamePart]:
        for p in self.parts:
            yield from p.name.parts

    def __post_init__(self) -> None:
        hash(self)
        check.not_empty(self.parts)
        for p in self.parts[:-1]:
            check.equal(p.kind, 'dir')

        pd: ta.Dict[DeployPathPlaceholder, ta.List[int]] = {}
        for i, np in enumerate(self.name_parts):
            if isinstance(np, PlaceholderDeployPathNamePart):
                pd.setdefault(ta.cast(DeployPathPlaceholder, np.placeholder), []).append(i)

        # if 'tag' in pd and 'app' not in pd:
        #     raise DeployPathError('Tag placeholder in path without app', self)

    @property
    def kind(self) -> ta.Literal['file', 'dir']:
        return self.parts[-1].kind

    def render(self, placeholders: ta.Optional[ta.Mapping[DeployPathPlaceholder, str]] = None) -> str:
        return ''.join([p.render(placeholders) for p in self.parts])

    @classmethod
    def parse(cls, s: str) -> 'DeployPath':
        check.non_empty_str(s)
        ps = split_keep_delimiter(s, '/')
        return cls(tuple(DeployPathPart.parse(p) for p in ps))



# ruff: noqa: UP006 UP007
from .apps import DeployAppManager
from .paths.manager import DeployPathsManager
from .specs import DeploySpec


class DeployManager:
    def __init__(
            self,
            *,
            apps: DeployAppManager,
            paths: DeployPathsManager,
    ):
        super().__init__()

        self._apps = apps
        self._paths = paths

    async def run_deploy(
            self,
            spec: DeploySpec,
    ) -> None:
        self._paths.validate_deploy_paths()

        #

        await self._apps.prepare_app(spec)



# ruff: noqa: UP006 UP007
import abc
import dataclasses as dc
import hashlib
import typing as ta

from omlish.lite.cached import cached_nullary
from omlish.lite.check import check

from .types import DeployApp
from .types import DeployKey
from .types import DeployRev


##


def check_valid_deploy_spec_path(s: str) -> str:
    check.non_empty_str(s)
    for c in ['..', '//']:
        check.not_in(c, s)
    check.arg(not s.startswith('/'))
    return s


##


@dc.dataclass(frozen=True)
class DeployGitRepo:
    host: ta.Optional[str] = None
    username: ta.Optional[str] = None
    path: ta.Optional[str] = None

    def __post_init__(self) -> None:
        check.not_in('..', check.non_empty_str(self.host))
        check.not_in('.', check.non_empty_str(self.path))


@dc.dataclass(frozen=True)
class DeployGitSpec:
    repo: DeployGitRepo
    rev: DeployRev

    subtrees: ta.Optional[ta.Sequence[str]] = None

    def __post_init__(self) -> None:
        check.non_empty_str(self.rev)
        if self.subtrees is not None:
            for st in self.subtrees:
                check.non_empty_str(st)


##


@dc.dataclass(frozen=True)
class DeployVenvSpec:
    interp: ta.Optional[str] = None

    requirements_files: ta.Optional[ta.Sequence[str]] = None
    extra_dependencies: ta.Optional[ta.Sequence[str]] = None

    use_uv: bool = False


##


@dc.dataclass(frozen=True)
class DeployConfFile:
    path: str
    body: str

    def __post_init__(self) -> None:
        check_valid_deploy_spec_path(self.path)


#


@dc.dataclass(frozen=True)
class DeployConfLink(abc.ABC):  # noqa
    """
    May be either:
     - @conf(.ext)* - links a single file in root of app conf dir to conf/@conf/@dst(.ext)*
     - @conf/file - links a single file in a single subdir to conf/@conf/@dst--file
     - @conf/ - links a directory in root of app conf dir to conf/@conf/@dst/
    """

    src: str

    def __post_init__(self) -> None:
        check_valid_deploy_spec_path(self.src)
        if '/' in self.src:
            check.equal(self.src.count('/'), 1)


class AppDeployConfLink(DeployConfLink):
    pass


class TagDeployConfLink(DeployConfLink):
    pass


#


@dc.dataclass(frozen=True)
class DeployConfSpec:
    files: ta.Optional[ta.Sequence[DeployConfFile]] = None

    links: ta.Optional[ta.Sequence[DeployConfLink]] = None

    def __post_init__(self) -> None:
        if self.files:
            seen: ta.Set[str] = set()
            for f in self.files:
                check.not_in(f.path, seen)
                seen.add(f.path)


##


@dc.dataclass(frozen=True)
class DeploySpec:
    app: DeployApp

    git: DeployGitSpec

    venv: ta.Optional[DeployVenvSpec] = None

    conf: ta.Optional[DeployConfSpec] = None

    @cached_nullary
    def key(self) -> DeployKey:
        return DeployKey(hashlib.sha256(repr(self).encode('utf-8')).hexdigest()[:8])



# ruff: noqa: UP006 UP007
import os.path
import typing as ta

from omlish.lite.inject import InjectorBindingOrBindings
from omlish.lite.inject import InjectorBindings
from omlish.lite.inject import inj
from omlish.os.atomics import AtomicPathSwapping

from ..commands.inject import bind_command
from .apps import DeployAppManager
from .commands import DeployCommand
from .commands import DeployCommandExecutor
from .conf import DeployConfManager
from .config import DeployConfig
from .deploy import DeployManager
from .git import DeployGitManager
from .interp import InterpCommand
from .interp import InterpCommandExecutor
from .paths.inject import bind_deploy_paths
from .paths.owners import DeployPathOwner
from .tmp import DeployTmpManager
from .types import DeployHome
from .venvs import DeployVenvManager


def bind_deploy(
        *,
        deploy_config: DeployConfig,
) -> InjectorBindings:
    lst: ta.List[InjectorBindingOrBindings] = [
        inj.bind(deploy_config),

        bind_deploy_paths(),
    ]

    #

    def bind_manager(cls: type) -> InjectorBindings:
        return inj.as_bindings(
            inj.bind(cls, singleton=True),

            *([inj.bind(DeployPathOwner, to_key=cls, array=True)] if issubclass(cls, DeployPathOwner) else []),
        )

    #

    lst.extend([
        bind_manager(DeployAppManager),

        bind_manager(DeployConfManager),

        bind_manager(DeployGitManager),

        bind_manager(DeployManager),

        bind_manager(DeployTmpManager),
        inj.bind(AtomicPathSwapping, to_key=DeployTmpManager),

        bind_manager(DeployVenvManager),
    ])

    #

    lst.extend([
        bind_command(DeployCommand, DeployCommandExecutor),
        bind_command(InterpCommand, InterpCommandExecutor),
    ])

    #

    if (dh := deploy_config.deploy_home) is not None:
        dh = os.path.abspath(os.path.expanduser(dh))
        lst.append(inj.bind(dh, key=DeployHome))

    return inj.as_bindings(*lst)
